package org.kantega.jexmec.jarfiles;

/*
 * Copyright 2011 Kantega AS
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.*;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;

/**
 * Class loader which loads plugin classes, resources and dependencies from a jar file or directory. Jar files under <code>META-INF/lib/</code>
 * are also added to the classpath, allowing for plugin-local embedded dependencies.
 * This class loader will treat resources under <code>META-INF/services/<i>PluginClass</i>/</code> as <i>local</i>,
 * meaning that parent classloaders will not be considered for these resources. This will protect plugins in parent class loaders
 * from being loaded once for each child class loader.
 */
public class EmbeddedLibraryPluginClassLoader extends URLClassLoader {

    /**
     * Loader classes and resources from the given directory. Jar files under META-INF/lib are added to the classpath.
     * @param directory the directory to load classes and resources from
     * @param parent the parent class loader
     */
    public EmbeddedLibraryPluginClassLoader(File directory, ClassLoader parent) {
        this(directory, parent, null, false);
    }

    /**
     * Loads classes and resources from the given jar file. If the jar file embeds other jar files in the <code>META-INF/lib</code>
     * directory, these will be unpacked to the given work directory and then added to the classpath.
     * @param jarFile the jar file to load classes and resources from
     * @param parent the parent class loader
     * @param workDir the directory where embedded libraries can be unpacked
     */
    public EmbeddedLibraryPluginClassLoader(File jarFile, ClassLoader parent, File workDir) {
        this(jarFile, parent, workDir, true);
    }

    private EmbeddedLibraryPluginClassLoader(File location, ClassLoader parent, File workDir, boolean isJarFile) {
        super(getUrls(location, workDir, isJarFile), parent);


    }

    private static URL[] getUrls(File location, File workDir, boolean expectJarFile) {

        if(!location.exists()) {
            throw new IllegalArgumentException("No such file or directory " + location);
        }

        if(expectJarFile && location.isDirectory()) {
            throw new IllegalArgumentException("Expected " +location + " to be a jar file, not a directory.");
        }
        if(!expectJarFile && location.isFile()) {
            throw new IllegalArgumentException("Expected " +location + " to be a directory, not a file.");
        }
        
        List<URL> urls = new ArrayList<URL>();


        try {
            if (location.isDirectory()) {

                urls.add(location.toURI().toURL());
                File libDir = new File(location, "META-INF/lib");
                if (libDir.exists()) {
                    for (File lib : libDir.listFiles()) {

                        urls.add(lib.toURI().toURL());

                    }
                }

            } else {
                String pluginName = location.getName();
                pluginName = pluginName.contains(".") ? pluginName.substring(0, pluginName.lastIndexOf(".")) : pluginName;

                final File pluginWorkDir = new File(workDir, pluginName);


                urls.add(location.toURI().toURL());

                JarFile jarFile = new JarFile(location);
                List<JarEntry> entries = Collections.list(jarFile.entries());

                Set<String> libraries = new HashSet<String>();
                for (JarEntry entry : entries) {

                    if (entry.getName().startsWith("META-INF/lib") && !entry.isDirectory() && (entry.getName().endsWith(".jar"))) {
                        String depName = entry.getName().substring(entry.getName().lastIndexOf("/") + 1);

                        File expandTo = new File(pluginWorkDir, depName);


                        if (!expandTo.exists() || location.lastModified() > expandTo.lastModified()) {
                            expandTo.getParentFile().mkdirs();
                            expand(entry, jarFile, expandTo);

                        }
                        urls.add(expandTo.toURI().toURL());
                        libraries.add(depName);
                    }


                }

                // Delete stale dependencies
                if (pluginWorkDir.exists()) {
                    for (File file : pluginWorkDir.listFiles()) {
                        if (!libraries.contains(file.getName())) {
                            file.delete();
                        }
                    }
                }


            }
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
        return urls.toArray(new URL[urls.size()]);
    }

    private static void expand(JarEntry entry, JarFile jarFile, File expandTo) throws IOException {
        copyStreamToFile(jarFile.getInputStream(entry), expandTo);
    }

    private static void copyStreamToFile(InputStream input, File expandTo) throws IOException {
        FileOutputStream output = null;
        try {
            output = new FileOutputStream(expandTo);
            byte[] buffer = new byte[1024 * 4];
            int read = 0;
            while (-1 != (read = input.read(buffer))) {
                output.write(buffer, 0, read);
            }
        } finally {
            if (output != null) {
                output.close();
            }
        }

    }
}

